use std::{
	iter,
	env,
	fs::OpenOptions,
	collections::HashMap,
	str::FromStr,
	fs::File,
};
use memmap2::Mmap;
use rouille::{Request, Response, router};
use parsebtrfs::{
	btrfs_structs::{TreeID, Value::Extent, Value::BlockGroup, ParseError, NODE_SIZE, ItemType, Item, RootItem, Value, Key, ExtentDataBody},
	btrfs_lookup::Tree, key,
};

const COLORS: &[&str] = &["#e6194b", "#3cb44b", "#ffe119", "#4363d8", "#f58231", "#911eb4", "#46f0f0", "#f032e6", "#bcf60c", "#fabebe", "#008080", "#e6beff", "#9a6324", "#fffac8", "#800000", "#aaffc3", "#808000", "#ffd8b1", "#000075", "#808080", "#000000"];

fn main() -> Result<(), MainError> {
	let filename = env::args().skip(1).next().ok_or("Argument required")?;

	let file = OpenOptions::new().read(true).open(filename)?;
	let image = unsafe { Mmap::map(&file)? };

	const O_DIRECT: i32 = 0x4000;
//	let file = OpenOptions::new().read(true).custom_flags(O_DIRECT).open(filename)?;
//	let image = unsafe { MmapOptions::new().len(493921239040usize).map(&file)? };

//	return Ok(());

	/*
	let mystery_addr = 0x2f_2251_c000;
	let addr_map = AddressMap::new(&image)?;
	let mystery_addr_phys = addr_map.to_phys(mystery_addr).unwrap() as usize;
	let mystery_node = Node::parse(&image[mystery_addr_phys .. ])?;

	println!("{:#x?}", &mystery_node);
*/

	rouille::start_server("127.0.0.1:8080", move |request| {
		router!(
			request,
			(GET) ["/"] => http_main_boxes(&image, request),
			(GET) ["/root"] => http_root(&image, None, request),
			(GET) ["/tree/{tree}", tree: String] => http_tree(&image, &tree, None, request),
			(GET) ["/tree/{tree}/{key}", tree: String, key: String] => http_tree(&image, &tree, Some(&key), request),
			(GET) ["/favicon.ico"] => Response::empty_404(),
			(GET) ["/style.css"] => Response::from_file("text/css", File::open("style.css").unwrap()),
			_ => Response::empty_404(),
		)
	});
}

static CIRCLE_IMAGE: &str =
	"data:image/png;base64,\
	 iVBORw0KGgoAAAANSUhEUgAAAAoAAAAKCAYAAACNMs+9AAAAP0lEQVQY02NgoBn4//+//P///yf9\
	 ////DRRP+v//vzw2hZP+Y4JJ2BS+waLwDUyeiVinIStchkV+GfmeoRoAAJqLWnEf4UboAAAAAElF\
	 TkSuQmCC";

static HTML_HEADER: &str = r###"<html>
<head>
<link rel="stylesheet" href="/style.css">
<script>
document.addEventListener("DOMContentLoaded", function() {
  folds = document.getElementsByClassName("fold");
  for(i = 0; i < folds.length; i++) {
//    folds[i].classList.add("open");
    folds[i].addEventListener("click", (event) => {
      folds[i].classList.add("open");
    });
  }
});
</script>
</head>
<body>
"###;

static HTML_FOOTER: &str = r###"</body>
</html>
"###;

static EXPLANATION_TEXT: &str = "\
<h3>Chunks</h3>
<p>On the highest level, btrfs splits the disk into <b>chunks</b> (also called <b>block groups</b>). They can have different sizes, with 1GiB being typical in a large file system. Each chunk can either contain data or metadata.<p>

<p>Here we look at the metadata chunks. They contain the B-treesm which btrfs gets its name from. They are key-value stores for different kinds of information. For example, the filesystem tree stores which files and directories are in the filesystem, and the extent tree stores which areas of the disk are in use. Each B-tree consists of a number of 16KiB <b>nodes</b>, here symbolized by colorful boxes, with the color indicating which tree the node belongs to. Most of the nodes are <b>leaves</b>, which contain the actual key-value pairs. The others are <b>interior nodes</b>, and we indicate them with a little white circle. They are important to find the leaf a key is stored in.</p>";

fn root_key_desc(id: u64) -> Option<&'static str> {
	match id {
		1 => Some("root"),
		2 => Some("extent"),
		3 => Some("chunk"),
		4 => Some("device"),
		5 => Some("filesystem"),
		6 => Some("root directory"),
		7 => Some("checksum"),
		8 => Some("quota"),
		9 => Some("UUID"),
		10 => Some("free space"),
		11 => Some("block group"),
		0xffff_ffff_ffff_fff7 => Some("data reloc"),
		_ => None,
	}
}

fn http_main_boxes(image: &[u8], _req: &Request) -> Response {
	let mut treecolors: HashMap<u64, &str> = HashMap::new();

	let mut result = String::new();

	let explanation_tablerowformat = |c: &str, t: &str| format!(
		"<tr>\
		 <td><table><tr><td style=\"height:10px;width:10px;padding:0;background:{};\"></td></tr></table></td>\
		 <td><table><tr><td style=\"height:10px;width:10px;padding:0;background:{};\"><img src=\"{}\" /></td></tr></table></td>\
		 <td>{}</td>\
		 </tr>\n",
		c, c, CIRCLE_IMAGE, t);
	let explanation_tablerowformat_leafonly = |c,t| format!(
		"<tr>\
		 <td><table><tr><td style=\"height:10px;width:10px;padding:0;background:{};\"></td></tr></table></td>\
		 <td></td>\
		 <td>{}</td>\
		 </tr>\n",
		c, t);

	let cellformat = |c| format!(
		"<td style=\"height:10px;width:10px;padding:0;background:{};\"></td>\n",
		c);
	let cellformat_higher = |c,_| format!(
		"<td style=\"height:10px;width:10px;padding:0;background:{}\"><img src=\"{}\" /></td>\n",
		c, CIRCLE_IMAGE);

	result.push_str(&"<details>\n<summary>What am I seeing here?</summary>");
	result.push_str(EXPLANATION_TEXT);

	// tree explanations
	result.push_str(&"<table style=\"margin: 0 auto;\">\n");
	result.push_str(&explanation_tablerowformat_leafonly("lightgrey", "unused or outdated node"));
	treecolors.insert(1, COLORS[treecolors.len() % COLORS.len()]);
	result.push_str(&explanation_tablerowformat(treecolors[&1], "root tree"));

	treecolors.insert(3, COLORS[treecolors.len() % COLORS.len()]);
	result.push_str(&explanation_tablerowformat(treecolors[&3], "chunk tree"));

	let roots = Tree::root(image).unwrap();
	for item in roots.iter() {
		if item.key.key_type == ItemType::Root {
			let treedesc: String = match &item.key.key_id {
				1 => format!("root tree"),
				2 => format!("extent tree"),
				3 => format!("chunk tree"),
				4 => format!("device tree"),
				5 => format!("filesystem tree"),
				6 => format!("root directory"),
				7 => format!("checksum tree"),
				8 => format!("quota tree"),
				9 => format!("UUID tree"),
				10 => format!("free space tree"),
				11 => format!("block group tree"),
				0xffff_ffff_ffff_fff7 => format!("data reloc tree"),
				x @ 0x100 ..= 0xffff_ffff_ffff_feff => format!("file tree, id = {}", x),
				x => format!("other tree, id = {}", x),
			};

			treecolors.insert(item.key.key_id, COLORS[treecolors.len() % COLORS.len()]);
			result.push_str(&explanation_tablerowformat(
				treecolors[&item.key.key_id],
				&treedesc
			));
		}
	}
	result.push_str(&"</table>\n");
	result.push_str(&"</details>\n");

	let extent_tree = Tree::new(&image, TreeID::Extent).unwrap();
	let mut extent_tree_iterator = extent_tree.iter();

	// current_blockgroup == None: haven't encountered a blockgroup yet
	// metadata_items == None: current blockgroup is not metadata or system
	let mut current_blockgroup = None;
	let mut metadata_items: Option<Vec<Option<(u64, u64)>>> = None;

	let metadata_blockgroups = iter::from_fn(|| {
		while let Some(item) = extent_tree_iterator.next() {
//			println!("Got key: {:x?}", &item.key);
			match &item.value {
				BlockGroup(bg) => {
					println!("{:x?}", item.key);
					let result = (current_blockgroup.take(), metadata_items.take());

					let nodes_in_blockgroup = item.key.key_offset as usize / NODE_SIZE;
					if bg.flags & 0x01 == 0 {
						metadata_items = Some(vec![None; nodes_in_blockgroup]);
					} else {
						metadata_items = None;
					}
					current_blockgroup = Some(item);

					if let (Some(bg), met) = result {
						return Some((bg, met));
					}
				},
				Extent(e) => {
					if let Some(bg_item) = &current_blockgroup {
						if let Some(met) = &mut metadata_items {
							let bg_start = bg_item.key.key_id;
							let node_addr = item.key.key_id;
							let tree_id = e.block_refs.iter().filter(|&(t,_)|t == &ItemType::TreeBlockRef).count() as u64;
							let index = (node_addr - bg_start) as usize / NODE_SIZE;
							if index < met.len() {
								met[index] = Some((tree_id, item.key.key_offset));
							} else {
								println!("Warning: extent out of block group range: {:x?}", &item.key);
							}
						}
					} else {
						println!("Warning: extent without matching block group: {:x?}", &item.key);
					}
				},
				_ => {},//panic!("Unexpected item in extent tree: {:x?}", item.key)
			}
		}

		let result = (current_blockgroup.take(), metadata_items.take());
		if let (Some(bg), met) = result {
			return Some((bg, met));
		} else {
			return None;
		}
	});

	let mut last_key = 0;

	// colorful table
	for (bg, nodes) in metadata_blockgroups {
		if bg.key.key_id < last_key {
			println!("Error: going backwards!");
			break;
		} else {
			last_key = bg.key.key_id;
		}

		let bg_value = match &bg.value {
			BlockGroup(bgv) => bgv,
			_ => panic!("Expected BlockGroup value"),
		};

		// header
		result.push_str(
			&format!(
				"<h3 style=\"text-align: center;\">{:x} - {:x} ({}, {})</h3><p>Physical: {}</p>\n",
				bg.key.key_id,
				bg.key.key_id + bg.key.key_offset,
				match bg.key.key_offset {
					x if x <= (1<<11) => format!("{} B", x),
					x if x <= (1<<21) => format!("{} KiB", x as f64 / (1u64<<10) as f64),
					x if x <= (1<<31) => format!("{} MiB", x as f64 / (1u64<<20) as f64),
					x if x <= (1<<41) => format!("{} GiB", x as f64 / (1u64<<30) as f64),
					x if x <= (1<<51) => format!("{} TiB", x as f64 / (1u64<<40) as f64),
					x @             _ => format!("{} PiB", x as f64 / (1u64<<50) as f64),
				},
				match bg_value.flags & 0x07 {
					0x01 => "Data",
					0x02 => "System",
					0x04 => "Metadata",
					_ => "???",
				},
				match extent_tree.addr_map.as_ref().0.binary_search_by_key(&bg.key.key_id, |x|x.0) {
					Ok(i) => format!("{:x?}", &extent_tree.addr_map.as_ref().0[i].2),
					_ => String::from(""),
				}
			)
		);

		if let Some(nodes) = nodes {
			result.push_str("<table style=\"margin: 0 auto;\">\n<tr>\n");

			for (i, &n) in nodes.iter().enumerate() {
				if i % 64 == 0 && i != 0 {
					result.push_str("</tr>\n<tr>\n");
				}

				if let Some((tid, level)) = n {
					let color: Option<&str> = treecolors.get(&tid).map(|x|*x);
					let color = color.unwrap_or_else(|| {
						println!("Unknown color for id: {}", &tid);
						let color: &str = COLORS[treecolors.len() % COLORS.len()];
						treecolors.insert(tid, color);
						color
					});
					if level == 0 {
						result.push_str(&cellformat(color));
					} else {
						result.push_str(&cellformat_higher(color, level));
					}
				} else {
					result.push_str(&cellformat("lightgrey"));
				}
			}

			result.push_str("</tr>\n</table>\n");
		}
	}

	Response::html(result)
}

fn html_item_row(key: Key, value: String, index: usize) -> String {
	let even = if index % 2 == 0 { "even" } else { "odd" };
	format!("<tr class = \"{even}\"><td>{:X}</td><td>{:?}</td><td>{:X}</td><td>{}</td></tr>\n",
			key.key_id, key.key_type, key.key_offset, value)
}

fn html_item_row_id_desc(key: Key, value: String, index: usize, desc: &str) -> String {
	let even = if index % 2 == 0 { "even" } else { "odd" };
	format!("<tr class = \"{even}\"><td>{}</td><td>{:?}</td><td>{:X}</td><td>{}</td></tr>\n",
			desc, key.key_type, key.key_offset, value)
}

fn item_value_string(tree_id: u64, key: Key, val: &Value) -> String {
	match val {
		Value::Root(_) => {
			format!("<a href=\"/tree/{}\">go</a>", key.key_id)
		},
		Value::Dir(dir_item) =>
			format!("name: {:?}, location: <a href=\"/tree/{}/{:x}\">{:X}<a>",
					&dir_item.name,
					tree_id,
					dir_item.location.key_id,
					dir_item.location.key_id),
		Value::DirIndex(dir_item) =>
			format!("name: {:?}, location: <a href=\"/tree/{}/{:x}\">{:X}<a>",
					&dir_item.name,
					tree_id,
					dir_item.location.key_id,
					dir_item.location.key_id),
		Value::Inode(inode_item) =>
			format!("mode: {:o}, ctime: {}, mtime: {}, otime: {}",
					inode_item.mode,
					inode_item.ctime.sec,
					inode_item.mtime.sec,
					inode_item.otime.sec),
		Value::ExtentData(extent_data_item) =>
			match &extent_data_item.data {
				ExtentDataBody::Inline(data) => format!("inline, len: {}", data.len()),
				ExtentDataBody::External(ext_extent) =>
					format!("external, disk_bytenr: {}, disk_num_bytes: {}, offset: {}, num_bytes: {}",
							ext_extent.disk_bytenr,
							ext_extent.disk_num_bytes,
							ext_extent.offset,
							ext_extent.num_bytes),
			},
		Value::Ref(ref_item) =>
			format!("name: {:?}, index: {}",
					&ref_item.name,
					ref_item.index),
		Value::Extent(extent_item) =>
			format!("flags: {}, block_refs: {:?}", extent_item.flags, extent_item.block_refs),
		Value::BlockGroup(blockgroup_item) =>
			format!("used: {} bytes", blockgroup_item.used),
		Value::DevExtent(dev_extent_item) =>
			format!("chunk_tree: {}, chunk_offset: {:x}, length: {} bytes", dev_extent_item.chunk_tree, dev_extent_item.chunk_offset, dev_extent_item.length),
		Value::UUIDSubvol(uuid_subvol_item) =>
			format!("subvolume id: {}", uuid_subvol_item.subvol_id),
		Value::FreeSpaceInfo(free_space_info) =>
			format!("extent_count: {}, flags: {}", free_space_info.extent_count, free_space_info.flags),
		Value::Dev(dev_item) =>
			format!("total_bytes: {}", dev_item.total_bytes),
		Value::Chunk(chunk_item) =>
			format!("size: {}", chunk_item.size),
		_ => {
			println!("{:?}", val);
			String::new()
		},
	}
}

fn http_tree_internal(tree: &Tree, tree_id: u64, key: Option<u64>) -> Response {
	let mut table: Vec<String> = Vec::new();

	table.push(String::from("<table><tbody>"));

	let k = Key {key_id: key.unwrap_or(0), key_type: ItemType::Invalid, key_offset: 0};
	let mut last_key_id: Option<u64> = None;
	let mut key_index = 0;

	for item in tree.range(k..).take(3000) {
		if last_key_id.replace(item.key.key_id).filter(|x| *x != item.key.key_id).is_some() {
			key_index += 1;
		}

		match &item.value {
			Value::Root(_) => {
				let id = match root_key_desc(item.key.key_id) {
					Some(desc) => format!("{:X} ({})", item.key.key_id, desc),
					None => format!("{:X}", item.key.key_id),
				};
				let value = item_value_string(tree_id, item.key, &item.value);
				table.push(html_item_row_id_desc(item.key, value, key_index, &id));
			},
			_ => {
				table.push(html_item_row(item.key, item_value_string(tree_id, item.key, &item.value), key_index));
			},
		}
	}

	table.push(String::from("</tbody></table>"));

	let mut result = String::new();

	result.push_str(HTML_HEADER);
	result.extend(table.iter().map(<String as AsRef<str>>::as_ref));
	result.push_str(HTML_FOOTER);

	Response::html(result)

}

fn http_tree(image: &[u8], tree_id: &str, key: Option<&str>, _req: &Request) -> Response {
	let key_id = key.and_then(|x|u64::from_str_radix(x, 16).ok());
	let tree_id = u64::from_str(tree_id).unwrap();
	let tree = if tree_id == 1 {
		Tree::root(image).unwrap()
	} else if tree_id == 3 {
		Tree::chunk(image).unwrap()
	} else {
		Tree::new(image, tree_id).unwrap()
	};
	http_tree_internal(&tree, tree_id, key_id)
}

fn http_root(image: &[u8], key: Option<&str>, _req: &Request) -> Response {
	let key_id = key.and_then(|x|u64::from_str_radix(x, 16).ok());
	let tree = Tree::root(image).unwrap();
	http_tree_internal(&tree, 1, key_id)
}


// ----- Error handling -----

pub struct MainError(String);

impl std::error::Error for MainError {}

impl std::fmt::Debug for MainError {
	fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
		write!(f, "{}", &self.0)
	}
}

impl std::fmt::Display for MainError {
	fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
		write!(f, "{}", &self.0)
	}
}

impl From<String> for MainError {
	fn from(value: String) -> MainError {
		MainError(value)
	}
}

impl From<&str> for MainError {
	fn from(value: &str) -> MainError {
		MainError::from(String::from(value))
	}
}

impl From<ParseError> for MainError {
	fn from(value: ParseError) -> MainError {
		MainError::from(format!("BTRFS format error: {value}"))
	}
}

impl From<std::io::Error> for MainError {
	fn from(value: std::io::Error) -> MainError {
		MainError::from(format!("IO error: {value}"))
	}
}

/*
fn main() -> Result<(), std::io::Error> {
	let file = File::open("../image")?;
	let image = unsafe { Mmap::map(&file)? };

	let addr = AddressTranslation::new(&image);

	rouille::start_server("127.0.0.1:8080", move |request| {
		http_main_list(&image, &addr, request)
	});
}


fn http_main_list(image: &[u8], addr: &AddressTranslation, req: &Request) -> Response {
	let chunk_offset = 0x02500000;
	let nodes_in_chunk = 2048;
	let mut result = String::new();

	result.push_str("<body>\n");

	for i in 0..nodes_in_chunk {
		let node = read_node(&image, chunk_offset + i*0x4000);

		let active = ACTIVE_NODES.contains(&(i*0x4000));
		let style = if active { "color:black;" } else { "color:lightgray;" };

		let newline = format!("<p style=\"{}\">{:x} {} {} {}\n<ul>\n",
							  style,
							  chunk_offset + i*0x4000,
							  node.level,
							  node.items.len(),
							  node.generation);
		result.push_str(&newline);

		for item in &node.items {
			let newline = format!("<li style=\"{}\">{:016x} {:?} {:x}</li>\n",
								  style,
								  item.key.key_id,
								  item.key.key_type,
								  item.key.key_offset);
			result.push_str(&newline);
		}

		result.push_str("</ul></p>\n");
	}

	Response::html(result)
}
*/

/*
fn read_node_log(image: &[u8], trans: &AddressTranslation, log: u64) -> Option<Box<BtrfsNode>> {
	let phys = trans.to_phys(log)?;
	Some(read_node(image, phys as usize))
}

fn read_node(image: &[u8], offset: usize) -> Box<BtrfsNode> {
	let mut result = Box::new(BtrfsNode {
		csum:           FromBytes::get(image, offset),
		fs_uid:         FromBytes::get(image, offset + 0x20),
		bytenr:         FromBytes::get(image, offset + 0x30),
		flags:          FromBytes::get(image, offset + 0x38),
		chunk_tree_uid: FromBytes::get(image, offset + 0x40),
		generation:     FromBytes::get(image, offset + 0x50),
		owner:          FromBytes::get(image, offset + 0x58),
		nritems:        FromBytes::get(image, offset + 0x60),
		level:          FromBytes::get(image, offset + 0x64),
		items: Vec::new(),
	});

	// assuming leaf for now

	for i in 0..result.nritems as usize {
		let key_id: u64 =       FromBytes::get(image, offset + 0x65 + i*0x19);
		let key_type_code: u8 = FromBytes::get(image, offset + 0x65 + i*0x19 + 0x08);
		let key_offset: u64 =   FromBytes::get(image, offset + 0x65 + i*0x19 + 0x09);
		let data_offset: u32 =  FromBytes::get(image, offset + 0x65 + i*0x19 + 0x11);
		let data_size: u32 =    FromBytes::get(image, offset + 0x65 + i*0x19 + 0x15);

		let key_type = itemtype_from_code(key_type_code);
		let data_slice = &image[(offset + 0x65 + data_offset as usize) .. (offset + 0x65 + data_offset as usize + data_size as usize)];

		let value = match key_type {
			BtrfsItemType::BlockGroup => BtrfsValue::BlockGroup(FromBytes::get(data_slice, 0)),
			BtrfsItemType::Metadata => BtrfsValue::Extent(FromBytes::get(data_slice, 0)),
			BtrfsItemType::Chunk => BtrfsValue::Chunk(FromBytes::get(data_slice, 0)),
			BtrfsItemType::Root => BtrfsValue::Root(FromBytes::get(data_slice, 0)),
			_ => BtrfsValue::Unknown(Vec::from(data_slice)),
		};

		result.items.push(BtrfsItem {
			key: BtrfsKey {
				key_id: key_id,
				key_type: key_type,
				key_offset: key_offset,
			},
			value: value,
		});
	}

	result
}
*/
